//     =========================== AVM_drawing.h ============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 17.04.2009
//     **********************************************************************

#if !defined(_AVM_drawing_H_INCLUDED_)
#define _AVM_drawing_H_INCLUDED_

#include <stdio.h>
#include "highgui.h"


inline CvScalar cvHLStoRGB(float h, float l, float s) {
	CvScalar Res;
	float b, g, r;

	if( s == 0 )
		b = g = r = l;
	else
	{
		static const int sector_data[][3]=
			{{1,3,0}, {1,0,2}, {3,0,1}, {0,2,1}, {0,1,3}, {2,1,0}};
		float tab[4];
		int sector;
    
		float p2 = l <= 0.5f ? l*(1 + s) : l + s - l*s;
		float p1 = 2*l - p2;

		h *= 0.016666666666666666f; // h /= 60;
		if( h < 0 )
			do h += 6; while( h < 0 );
		else if( h >= 6 )
			do h -= 6; while( h >= 6 );

		assert( 0 <= h && h < 6 );
		sector = cvFloor(h);
		h -= sector;

		tab[0] = p2;
		tab[1] = p1;
		tab[2] = p1 + (p2 - p1)*(1-h);
		tab[3] = p1 + (p2 - p1)*h;

		b = tab[sector_data[sector][0]];
		g = tab[sector_data[sector][1]];
		r = tab[sector_data[sector][2]];
	}
	
	Res.val[0] = uchar(b*255.0f);
	Res.val[1] = uchar(g*255.0f);
	Res.val[2] = uchar(r*255.0f);
	Res.val[3] = 0;

	return Res;
}


// Drawing of associative tree
class CvDrawAssociativeTree {

#define cAT_DiagramTotal 3
#define cAT_xBorderSize  25
#define cAT_yBorderSize  40
#define cAT_xThickness   2
#define cAT_yThickness   9
#define cAT_WidthMin     360
#define cAT_WidthMax     1024
#define cAT_WinName      "Associative tree"

	CvAssociativeMemory32S* am;

public:
	
	// Class constructor
	CvDrawAssociativeTree(CvAssociativeMemory32S* apAM) {
		am = apAM;
		cvNamedWindow(cAT_WinName, 1);
	}

	// Class destructor
	~CvDrawAssociativeTree() { cvDestroyWindow(cAT_WinName); }
	
	typedef enum {
		tHits = 0,
		tSearching,
		tActivity
	} AT_Type;


	struct TBasePoint {
		int     x;
		uint64	HitTotal;     // Total number of hitting
		double  ActualTick;   // Actual tick number
		double  LastTickDiff; // Last tick difference after searching
	};

	struct TParamStat {
		uint64      HitMax;
		double      TickMax;
		double      TickDiffAvr;
		double      TickDiffMax;
		double      StepsAvr;
		long        StepsMax;
	};

	// Data transmission structure for "doCountBottomBases" function
	struct TCountBottomBasesData {
		int Counter;
		int Level;
		int xBorderSize;
		TBasePoint*	BasePoint;
		TParamStat  s;
	};

	static bool doCountBottomBases(CvSize, short*, uint64, CvClusterStat* apStat,
													int, short  aLevel, void* args) {
		
		TCountBottomBasesData* d = (TCountBottomBasesData*)args;

		if(d->s.HitMax  < apStat->HitTotal)   { d->s.HitMax  = apStat->HitTotal; }
		if(d->s.TickMax < apStat->ActualTick) { d->s.TickMax = apStat->ActualTick; }
		if(d->s.TickDiffMax < apStat->LastTickDiff) { d->s.TickDiffMax = apStat->LastTickDiff; }
		if(d->s.StepsMax < apStat->StepsForSearch)  { d->s.StepsMax = apStat->StepsForSearch; }

		d->s.TickDiffAvr = (d->s.TickDiffAvr + apStat->LastTickDiff)/2.0;
		d->s.StepsAvr = (d->s.StepsAvr + apStat->StepsForSearch)/2.0;

		if(aLevel == d->Level) {
			d->BasePoint[d->Counter].HitTotal = apStat->HitTotal;
			d->BasePoint[d->Counter].ActualTick = apStat->ActualTick;
			d->BasePoint[d->Counter].LastTickDiff = apStat->LastTickDiff;
			d->BasePoint[d->Counter].x = d->xBorderSize + (d->Counter++)*cAT_xThickness;
		}

		return true;
	}

	// Data transmission structure for "doDrawBases" function
	struct TDrawBasesData {
		int Level;
		int ActIdx;
		int WrtIdx;
		int Ly;
		int Lh;
		double TickFrequency;
		IplImage*   TreeImage;
		AT_Type     dType;
		TBasePoint*	BasePoint;
		TParamStat  s;
	};

	static bool doDrawBases(CvSize, short*, uint64, CvClusterStat* apStat,
					  int aContainBasesTotal, short  aLevel, void* args) {
		
		TDrawBasesData* d = (TDrawBasesData*)args;

		int i, x = 0, xSum = 0;	float cHSum = 0;
		int bx  = d->BasePoint[d->ActIdx].x;

		if(aLevel == d->Level) {

			for(i=0; i<aContainBasesTotal; i++) {
				double ExpArg = 0;
				switch(d->dType) {
					case tHits:
						ExpArg = (d->s.HitMax)?(double(__int64(d->BasePoint[d->ActIdx].HitTotal<<5))/
												double(__int64(d->s.HitMax))):1.0; break;
					case tSearching:
						ExpArg = (d->s.TickDiffMax)?(((d->BasePoint[d->ActIdx].LastTickDiff*2.0)/
								  d->s.TickDiffMax)):1.0; break;

					case tActivity:
						double TickDiff = d->s.TickMax - d->BasePoint[d->ActIdx].ActualTick;
						ExpArg = (TickDiff)?((d->TickFrequency*5e+6)/TickDiff):1.0; break;
				}

				float cH = float(240.0 + 180.0*(1.0 - exp(-ExpArg - 0.1))); cHSum += cH;

				x = d->BasePoint[d->ActIdx++].x;

				CvPoint P1 = cvPoint(x, d->Ly);
				CvPoint P2 = cvPoint(x, d->Ly + d->Lh);

				if(d->dType == tSearching && d->BasePoint[d->ActIdx].LastTickDiff == d->s.TickDiffMax) {
					cvLine(d->TreeImage, P1, P2, CV_RGB(0,0xFF,0),2,8,0);
				} else {
					cvLine(d->TreeImage, P1, P2, cvHLStoRGB(cH,0.5f,1),1,8,0);
				}
				xSum += x;
			}

			cvLine(d->TreeImage, cvPoint(bx, d->Ly), cvPoint(x, d->Ly),
					cvHLStoRGB((aContainBasesTotal)?(cHSum/aContainBasesTotal):0,0.5f,1),1,8,0);
			
			switch(d->dType) {
				case tHits:      d->BasePoint[d->WrtIdx].HitTotal = apStat->HitTotal; break;
				case tSearching: d->BasePoint[d->WrtIdx].LastTickDiff = apStat->LastTickDiff; break;
				case tActivity:  d->BasePoint[d->WrtIdx].ActualTick = apStat->ActualTick; break;
			}
			d->BasePoint[d->WrtIdx++].x = (aContainBasesTotal)?(xSum/aContainBasesTotal):xSum;
		}

		return true;
	}

	void ShowAssociativeTree() {
		char* AT_Title[cAT_DiagramTotal] = {"Associative tree: Hits", "Searching", "Activity"};
		int i,d,lCount;
		char OutStr[256];
		CvFont BigFont;
		cvInitFont( &BigFont, 2, 0.8, 0.8, 0, 2, 8 );
		CvFont MiddleFont;
		cvInitFont( &MiddleFont, 2, 0.7, 0.7, 0.3, 1, 8 );
		CvFont SmallFont;
		cvInitFont( &SmallFont, 1, 1.0, 1.0, 0, 1, 8 );

		TBasePoint*	pBasePoint[cAT_DiagramTotal];
		int BasePointBufSize = am->GetTotalABases()*sizeof(TBasePoint);
		
		for(i=0; i<cAT_DiagramTotal; i++) {
			pBasePoint[i] = (TBasePoint*)cvAlloc(BasePointBufSize);
		}
		
		TCountBottomBasesData BBD;
		memset(&BBD, 0, sizeof(TCountBottomBasesData));

		int PreviouslyWidth = am->GetTotalABases()*cAT_xThickness + cAT_xBorderSize*2;

		BBD.xBorderSize = (PreviouslyWidth > cAT_WidthMax)?
			((cAT_xBorderSize*PreviouslyWidth)/cAT_WidthMax):cAT_xBorderSize;
		BBD.Level = am->GetNodeTotalLevels();
		BBD.BasePoint = pBasePoint[0];
		memset(pBasePoint[0], 0, BasePointBufSize);

		short DrawLevel = BBD.Level-1;

		am->DoForEachNode(doCountBottomBases, &BBD);

		memcpy(pBasePoint[1], pBasePoint[0], BasePointBufSize);
		memcpy(pBasePoint[2], pBasePoint[0], BasePointBufSize);

		int DiagramHeight = DrawLevel*cAT_yThickness + cAT_yBorderSize*2;
		
		CvSize Size;
		Size.width  = BBD.Counter*cAT_xThickness + BBD.xBorderSize*2;
		Size.height = (DiagramHeight)*cAT_DiagramTotal;

		if(Size.width < cAT_WidthMin) { Size.width = cAT_WidthMin; }

		TDrawBasesData DBD;

		DBD.s = BBD.s;
		DBD.TickFrequency = cvGetTickFrequency();
		DBD.TreeImage  = cvCreateImage(Size, IPL_DEPTH_8U, 3);
		cvSet(DBD.TreeImage,cvScalarAll(180),0);

		if(Size.width > cAT_WidthMax) { Size.width = cAT_WidthMax; }

		IplImage* DstImg = cvCreateImage(Size, IPL_DEPTH_8U, 3);
		cvSet(DstImg,cvScalarAll(180),0);

		int DiagramXBegin    = cAT_xBorderSize;
		int DiagramXEnd      = Size.width - cAT_xBorderSize;

		for(d=0; d<cAT_DiagramTotal; d++) {

			lCount = am->GetTotalLevels();
			for(i=DrawLevel; i; i--) {
				DBD.WrtIdx = 0;
				DBD.ActIdx = 0;
				DBD.Level  = i;
				DBD.dType  = AT_Type(d);
				DBD.Ly     = d*DiagramHeight + cAT_yBorderSize + (i-1)*cAT_yThickness;
				DBD.Lh     = cAT_yThickness;
				DBD.BasePoint = pBasePoint[d];

				am->DoForEachNode(doDrawBases, &DBD);
			}
		}

		cvResize(DBD.TreeImage, DstImg, CV_INTER_AREA);

		for(d=0; d<cAT_DiagramTotal; d++) {

			int DiagramYBase     = d*DiagramHeight;
			int DiagramYTop      = DiagramYBase + cAT_yBorderSize;
			int DiagramYBottom   = DiagramYBase + DiagramHeight - cAT_yBorderSize;
			int DiagramYBottomL1 = DiagramYBottom + 4;
			int DiagramYBottomT1 = DiagramYBottom + 20;

			cvPutText(DstImg, AT_Title[d], cvPoint(DiagramXBegin, DiagramYTop - 20), &MiddleFont, CV_RGB(0,0,0));

			switch(AT_Type(d)) {
				case tHits:      sprintf(OutStr, "Total bases/clusters: %-3lu/%-3lu",
											am->GetTotalABases(), am->GetClustersTotal()); break;
				case tSearching: sprintf(OutStr, "Time (mks): Avr[%4.1f] Max[%4.1f]",
									 DBD.s.TickDiffAvr/DBD.TickFrequency, DBD.s.TickDiffMax/DBD.TickFrequency); break;
				case tActivity:  sprintf(OutStr, "Steps: Avr[%.0f] Max[%u]", DBD.s.StepsAvr, DBD.s.StepsMax); break;
			}
			
			cvPutText(DstImg, OutStr, cvPoint(cAT_xBorderSize, DiagramYBottomT1),&SmallFont, CV_RGB(0,0,0));
			cvLine(DstImg, cvPoint(DiagramXBegin, DiagramYBottomL1),
						   cvPoint(DiagramXEnd, DiagramYBottomL1), cvScalarAll(0),2,8,0);

			lCount = am->GetTotalLevels();
			for(i=DrawLevel; i; i--) {
				if(am->GetBaseNodeLevel(lCount) == i) {
					int lY = DiagramYBase + 4 + cAT_yBorderSize + (i-1)*cAT_yThickness;
					cvLine(DstImg, cvPoint(DiagramXBegin-2, lY), cvPoint(DiagramXEnd, lY), cvScalarAll(100),1,CV_AA,0);

					sprintf(OutStr, "L%u", lCount);
					cvPutText(DstImg, OutStr, cvPoint(2, lY),&SmallFont, CV_RGB(0,0,0));
					cvPutText(DstImg, OutStr, cvPoint(DiagramXEnd+2, lY),&SmallFont, CV_RGB(0,0,0));
					lCount--;
				}
			}
		}

		cvShowImage(cAT_WinName, DstImg);

		for(i=0; i<cAT_DiagramTotal; i++) {	cvFree( &pBasePoint[i] ); }
		cvReleaseImage( &DBD.TreeImage );
		cvReleaseImage( &DstImg );
	}
};


#endif // !defined(_AVM_drawing_H_INCLUDED_)