//     ======================= AssociativeMemory.h ==========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 06.10.2009
//     **********************************************************************

#if !defined(_AssociativeMemory_H_INCLUDED_)
#define _AssociativeMemory_H_INCLUDED_

#include "cv.h"
#include "cxmisc.h"

// Definition of coordinate structure
struct CvXYZ32F {
	float x,y,z;
    // Set a coordinate
	void Set(float aX, float aY, float aZ) { x = aX; y = aY; z = aZ; }
};

// Definition of location structure
struct CvLocation {
	CvXYZ32F p; // Location point
	float   hA; // Horizontal survey angle
	float   vA; // Vertical survey angle
};

#define cAM_SimilarityUpTrd   0.54f  // Threshold of similarity on top
#define cAM_SimilarityDnTrd   0.45f  // Threshold of similarity on bottom

// Definition of parameter types
typedef enum {
	ptOptimizeLoop = 0,      // Period of optimization, ms (45000)
	ptTrackingDepth,         // Depth of tracking (40)
	ptTrcDist2Max,           // Threshold of square distance for neighbor objects
	ptExptectDistEr,         // Expected distance error for tracking
	ptRcgMxCellBase,         // Base cell value for recognition threshold calculation (20000)
	ptMaxClusterAge,         // Age of formed cluster (40)
	ptRcgMxExpRate_A,        // | Exponent rate of recognition matrix for learning (0.07)
	ptRcgMxExpRate_B,        // | (0.2)
	ptRcgMxExpRate_C,        // | (0.0015)
	ptSearchScaleStep,       // Scale step of searching  for recognition (0.25)
} CvAM_ParamType;

// Definition of the state constants for recognition function
typedef enum {
	cProcessingIsBegun = 0,
	cUnknownObject,
	cRecognizedObject,
	cGeneralizedObject,
	cTrackedObject,
	cLearnThisObject,
	cProcessingIsFinished,
	cDeletingTrackedObject
} CvAM_State;

// Parameters structure for recognition function
struct CvAM_Parameters {
	CvRect   ObjRect;            // Rectangle where object found
	double   Similarity;         // Similarity of the interest area to an object (0 ... 1)
	long     TotalABases;        // Total number of associative bases
	void**   ppUserTrackingInfo; // Pointer to the tracking info data written by user
};

// Structure of cluster statistics
struct CvClusterStat {
	uint64	       HitTotal;         // Total number of hitting
	double         ActualTick;       // Actual tick number
	double         LastTickDiff;     // Last tick difference after searching
	long           StepsForSearch;   // Steps needed for search 
};

// Definition of matrix cluster operation function
typedef bool (*TYPE_RcgMCFunc16S)(CvSize  aRcgMxSize, short* apRcgMx,
		                          uint64 aIndex, CvClusterStat* apStat,
								  int aContainBasesTotal, short aLevel, void* args);
typedef TYPE_RcgMCFunc16S CvRcgMCFunc16S;

#endif // !defined(_AssociativeMemory_H_INCLUDED_)
