//     ========================== ListInterface.h ===========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 10.11.2009
//     **********************************************************************

#if !defined(_ListInterface_H_INCLUDED_)
#define _ListInterface_H_INCLUDED_

//   MIN/MAX, NULL
#ifndef MIN
#define MIN(a,b)  ((a) > (b) ? (b) : (a))
#endif

#ifndef MAX
#define MAX(a,b)  ((a) < (b) ? (b) : (a))
#endif

#ifndef NULL
#define NULL 0
#endif

//========================================================================================
//================================== Class: iList ========================================
//========================================================================================

//   
template <class iData>
struct iListItem {
	iData d;          // 
	iListItem* next;  //     ( NULL   )
	iListItem* prev;  //    
};

//   
template <class iData>
class iList {
public:
	
	//   
	typedef iListItem<iData> iListItem;
	
private:
	
	int  mTotal; //    
	iListItem *mpFirst, *mpActual, *mpLast; //    
	
	//        
	inline void Reset() {
		mpFirst = NULL; mpActual = NULL; mpLast = NULL;
		mTotal = 0;
	}
	
public:
	
	//  
	iList() { Reset(); }
	
	//  
	~iList() { Clear(); }
	
	//  
	iList(iList<iData>& aList) {
		//  "" 
		mTotal   = aList.mTotal;
		mpFirst  = aList.mpFirst;
		mpActual = aList.mpActual;
		mpLast   = aList.mpLast;
		aList.Reset();
	}
	
	//   +=
	const iList& operator+=(const iList& aList) {
		iListItem *p = aList.mpFirst;
		while(p) { Append(p->d); p = p->next; }
		return *this;
	}
	
	//   
	const iList& operator=(const iList& aList) {
		Clear();
		*this += aList;
		return *this;
	}
	
	//    
	int Total() const { return mTotal; }
	
	//      
	iListItem* Append(const iData& aData) {
		
		iListItem* p = new iListItem;
		p->d = aData;
		p->next = NULL;
		if(!mpFirst) { mpFirst = p; }

		if(mpLast)   {
			p->prev = mpLast;
			mpLast->next = p;
		} else { p->prev = NULL; }
		
		mpLast = p; mTotal++;

		return p;
	}

	//        ( ) 
	iListItem* InsertBefore(const iData& aData) {
		if(!mpFirst) { return Append(aData); }

		iListItem* p = new iListItem;
		p->d = aData;

		if(mpActual) {
			if(mpActual->prev) { mpActual->prev->next = p; }
			p->prev = mpActual->prev; mpActual->prev = p;
			p->next = mpActual;
			if(mpActual == mpFirst) { mpFirst = p; }
		} else { p->prev = NULL; p->next = mpFirst; mpFirst = p; }

		mTotal++;

		return p;
	}

	//     
	iListItem* InsertBySorting(const iData& aData, iListItem* apStartFromPos = NULL) {
		if(!mpFirst) { return Append(aData); }

		iData* pData;

		if(apStartFromPos) {
			SetActualPosition(apStartFromPos);
		} else {
			GoAtStart();
		}

		while(pData = GetNext()) {
			if(*pData > aData) { return InsertBefore(aData); }
		}
		return Append(aData);
	}

	//     
	void GoAtStart() { mpActual = NULL; }

	//     
	void GoAtEnd() {  mpActual = mpLast; }

	//     
	iListItem* GetActualPosition() const { return mpActual; }

	//     
	void SetActualPosition(iListItem* apPosition) {	mpActual = apPosition; }

	//     
	void Write(const iData& aData) {
		if(mpActual) { mpActual->d = aData;	}
	}

	//     
	iData* Read() {
		if(mpActual) { return &mpActual->d;	} else { return NULL; }
	}

	//    
	iData* GetNext() {
		if(mpActual) {
			if(!mpActual->next) { return NULL; }

			mpActual = mpActual->next;
		} else { mpActual = mpFirst; }
		
		return &mpActual->d;
	}

	//    
	iData* GetPrev() {
		if(!mpActual) { return NULL; }

		iListItem *pActual = mpActual;
		mpActual = mpActual->prev;
		
		return &pActual->d;
	}

	//      
	iListItem* GetFirstPosition() const { return mpFirst; }

	//     
	static iData* GetNext(iListItem** appLI) {
		if(!appLI || !(*appLI)) { return NULL; }
		iData* p = &(*appLI)->d;
		*appLI = (*appLI)->next;
		return p;
	}
		
	//  
	void Clear() {
		iListItem *remove, *p = mpFirst;

		while(p) {	remove = p; p = p->next; delete remove; }
		Reset();
	}

	//   
	void Delete(iListItem* apDelItem) {
		if(apDelItem) {
			if(apDelItem == mpFirst) {
				
				if(apDelItem == mpLast) { mpLast = NULL; }

				mpFirst  = apDelItem->next;
				if(apDelItem->next) { apDelItem->next->prev = NULL; }
				
			} else {
				if(apDelItem == mpLast) {
					mpLast = apDelItem->prev;
					mpLast->next = NULL;
				} else {
					apDelItem->prev->next = apDelItem->next;
					apDelItem->next->prev = apDelItem->prev;
				}
			}
			delete apDelItem; mTotal--;
		}
	}

	//    
	void DeleteActual() {
		if(mpActual) {
			iListItem* pDelItem = mpActual;
			mpActual = mpActual->next;
			Delete(pDelItem);
		}
	}
};

//     
template <class iData, class iListItem>
bool ListsIsEqual(const iList<iData>& aL1, const iList<iData>& aL2, iListItem*) {
	if(aL1.Total() != aL2.Total()) { return false; }

	iListItem* pLi1 = aL1.GetFirstPosition();
	iListItem* pLi2 = aL2.GetFirstPosition();

	iData *pD1, *pD2;

	//   
	while(pD1 = iList<iData>::GetNext(&pLi1)) {
		pD2 = iList<iData>::GetNext(&pLi2);
		if(*pD1 != *pD2) { return false; }
	}

	return true;
}

//      iList
template <class iData>
bool operator==(const iList<iData>& aL1, const iList<iData>& aL2)
{
	return ListsIsEqual(aL1, aL2, (iList<iData>::iListItem*)NULL);
}

//      iList
template <class iData>
bool operator>(const iList<iData>& aL1, const iList<iData>& aL2)
{
	return (aL1.Total()>>1) > (aL2.Total()>>1);
}

//      iList
template <class iData>
bool operator<(const iList<iData>& aL1, const iList<iData>& aL2)
{
	return (aL1.Total()>>1) < (aL2.Total()>>1);
}

//========================================================================================
//============================== Class: iAssociatedList ==================================
//========================================================================================

//    
template <class iKey, class iData>
struct iAscData {
	iKey  Key;
	iData Data;
	void* pIdx;
};

//   
template <class iKey, class iData>
class iAssociatedList {
public:

	//    
	typedef iAscData<iKey, iData> iAscData;

	//   
	typedef iList<iAscData> iPairList;

	//     
	typedef iListItem<iAscData> iAscItem;

	//   "  "
	typedef void (*TYPE_IfKeyFound)(iPairList* apPL, const iAscData& aData, void* args);
	typedef TYPE_IfKeyFound iIfKeyFound;

private:

	#define cLI_IndexOrder 10 //   10 

	//    
	struct iIndexItem {
		iKey  Key;
		iAscItem *Pos;
	};

	//   
	typedef iList<iIndexItem> iIndexList;
	typedef iListItem<iIndexItem> iIndexListItem;

	double mKeyMin, mKeyMax; // /  
	iPairList*  mpPairList;  //    
	iIndexList* mpIndexList; //    
	
	
	//   
	typedef bool (*TYPE_Compare)(const iKey& k1, const iKey& k2, void* args);
	typedef TYPE_Compare iCompare;
	
	//    
	static bool doSimCompare(const iKey& k1, const iKey& k2, void* args) {
		double* pSimilarityThreshold = (double*)args;
		return iKey::Similarity(k1, k2) >= *pSimilarityThreshold;
	}
	
	//   
	static bool doIdenticalCompare(const iKey& k1, const iKey& k2, void*) {
		return k1 == k2;
	}
	
	//     
	iData* Find(const iKey& aKey, iCompare FuncCompare,
				double* apSimTrd = NULL, iIndexListItem* apStartFrom = NULL) {

		iAscData  *pLI;
		iIndexItem *pIdxDat;
		iData *pData = NULL;
		iPairList::iListItem *pFrom = NULL, *pTo = NULL;

		if(!mpPairList->Total()) { return NULL; }

		double KeyBase = mKeyMax - mKeyMin;

		if(aKey > mKeyMin && KeyBase) {
		
			double InKey = (aKey - mKeyMin)/KeyBase;
			double InKeyFrom = (apSimTrd)?(InKey + *apSimTrd - 1.0):InKey;
			double InKeyTo   = (apSimTrd)?(InKey - *apSimTrd + 1.0):InKey;

			//   
			if(apStartFrom) {
				mpIndexList->SetActualPosition(apStartFrom);
				pFrom = apStartFrom->d.Pos;
			} else {
				mpIndexList->GoAtStart();
			}

			while(pIdxDat = mpIndexList->GetNext()) {
				double ActKey = (pIdxDat->Key - mKeyMin)/KeyBase;
				if(InKeyTo < ActKey) {
					pTo = pIdxDat->Pos; break;
				} else {
					if(InKeyFrom > ActKey) { pFrom = pIdxDat->Pos; }
				}
			}

			//     
			mpPairList->SetActualPosition(pFrom);
		} else { mpPairList->GoAtStart(); }

		while((pLI = mpPairList->GetNext()) && mpPairList->GetActualPosition() != pTo) {
			//    ?
			if(FuncCompare(pLI->Key, aKey, apSimTrd)) {// Yes
				//  
				pData = &pLI->Data; break;
			}
		}

		return pData;
	}
	
public:

	//  
	iAssociatedList() {
		mpPairList  = new iPairList;
		mpIndexList = new iIndexList;
	}
	
	//  
	~iAssociatedList() {
		if(mpPairList)  { delete mpPairList;  mpPairList  = NULL; }
		if(mpIndexList) { delete mpIndexList; mpIndexList = NULL; }
	}
	
	//  
	iAssociatedList(iAssociatedList<iKey, iData>& aList) {
		// Copying an image of data
		mpPairList  = aList.mpPairList;  aList.mpPairList  = NULL;
		mpIndexList = aList.mpIndexList; aList.mpIndexList = NULL;
	}
	
	//   
	const iAssociatedList& operator=(const iAssociatedList& aList) {
		iAscData* pData;
		iAscItem* pLI = aList.mpPairList->GetFirstPosition();
		mpIndexList->Clear();
		mpPairList->Clear();

		for(int i=0; pData = iAssociatedList::GetNext(&pLI); i++) {
			double dKey = double(pData->Key);
			if(!i) {
				mKeyMin = dKey;
				mKeyMax = dKey;
			} else {
				if(mKeyMin > dKey) { mKeyMin = dKey; }
				if(mKeyMax < dKey) { mKeyMax = dKey; }
			}
			
			iAscItem* pNew = mpPairList->Append(*pData);

			if(!(i % cLI_IndexOrder)) {
				iIndexItem IdxDat;
				IdxDat.Key = pNew->d.Key;
				IdxDat.Pos = pNew;
				pNew->d.pIdx = mpIndexList->Append(IdxDat);
			} else {
				pNew->d.pIdx = NULL;
			}
		}

		return *this;
	}

	//  
	void AddPair(const iKey& aKey, const iData& aData,
				 iIfKeyFound FuncIfKeyFound = NULL, void* args = NULL) {

		iAscData d;
		d.Key  = aKey;
		d.Data = aData;
		d.pIdx = NULL;

		double dKey = double(aKey);
		if(!mpPairList->Total()) {
			mKeyMin = dKey;
			mKeyMax = dKey;
		} else {
			if(mKeyMin > dKey) { mKeyMin = dKey; }
			if(mKeyMax < dKey) { mKeyMax = dKey; }
		}

		iIndexItem *pIdxDat;
		iIndexListItem* pStartFrom = NULL;

		//   
		mpIndexList->GoAtStart();
		while(pIdxDat = mpIndexList->GetNext()) {
			if(pIdxDat->Key > aKey) { break; }
			pStartFrom = mpIndexList->GetActualPosition();
		}

		//  ?
		if(Find(aKey, doIdenticalCompare, NULL, pStartFrom)) {// 
			if(FuncIfKeyFound) {
				FuncIfKeyFound(mpPairList, d, args);
			} else {
				mpPairList->Write(d);
			}
			return;
		}

		//      
		iPairList::iListItem* pPL = mpPairList->InsertBySorting(d, pStartFrom?pStartFrom->d.Pos:NULL);

		//    ?
		if(!(mpPairList->Total() % cLI_IndexOrder) || mpPairList->Total() == 1) {// 
			iIndexItem IdxDat;
			IdxDat.Key = pPL->d.Key;
			IdxDat.Pos = pPL;
			pPL->d.pIdx = mpIndexList->Append(IdxDat);
		} else {
			pPL->d.pIdx = NULL;
		}
	}


	//  
	void DeletePair(const iKey& aKey) {
		if(Find(aKey, doIdenticalCompare)) {
			iAscItem* pFoundItem = mpPairList->GetActualPosition();

			if(pFoundItem->d.pIdx) {
				mpIndexList->Delete((iIndexListItem*)pFoundItem->d.pIdx);
			}

			mpPairList->DeleteActual();
		}
	}

	//         
	iData* Find(const iKey& aKey, double aSimilarityThreshold) {
		return Find(aKey, doSimCompare, &aSimilarityThreshold);
	}

	//        
	iData* Find(const iKey& aKey) {
		return Find(aKey, doIdenticalCompare);
	}

	//      
	iAscItem* GetFirstPosition() const { return mpPairList->GetFirstPosition(); }

	//    
	static iAscData* GetNext(iAscItem** appLI) {
		return iPairList::GetNext(appLI);
	}

	//     
	int Total() const { return mpPairList->Total(); }

	//  
	void Clear() {
		if(mpPairList)  { mpPairList->Clear(); }
		if(mpIndexList) { mpIndexList->Clear(); }
	}

	//     
	iPairList* GetPairListPtr() const { return mpPairList; }
};

//      iAscData<iKey, iData>
template <class iKey, class iData>
bool operator>(const iAscData<iKey, iData>& d1,
			   const iAscData<iKey, iData>& d2) {
	return d1.Key > d2.Key;
}

//       iAscData<iKey, iData>
template <class iKey, class iData>
bool operator!=(const iAscData<iKey, iData>& d1,
				const iAscData<iKey, iData>& d2)
{
	return d1.Key != d2.Key;
}

//      iAssociatedList
template <class iKey, class iData>
bool operator==(const iAssociatedList<iKey, iData>& aL1,
				const iAssociatedList<iKey, iData>& aL2)
{
	return ListsIsEqual(*aL1.GetPairListPtr(), *aL2.GetPairListPtr(),
						(iAssociatedList<iKey, iData>::iPairList::iListItem*)NULL);
}

//      iAssociatedList
template <class iKey, class iData>
bool operator>(const iAssociatedList<iKey, iData>& aL1,
			   const iAssociatedList<iKey, iData>& aL2)
{
	return aL1.GetPairListPtr()->Total() > aL2.GetPairListPtr()->Total();
}

//      iAssociatedList
template <class iKey, class iData>
bool operator<(const iAssociatedList<iKey, iData>& aL1,
			   const iAssociatedList<iKey, iData>& aL2)
{
	return aL1.GetPairListPtr()->Total() < aL2.GetPairListPtr()->Total();
}


//========================================================================================
//================================== Class: iIdString ====================================
//========================================================================================

//   
template <class iIdentifierType, class iIdentifierItem>
class iIdString {
	//    
	typedef iList<iIdentifierItem> iIdentifierList;

	//   
	typedef iAssociatedList<iIdentifierType, long> iString;

	long     mIdTotal;  //   
	iString* mpString;  //    

public:

	//     
	typedef iAscData<iIdentifierType, long> iIdData;

	//   
	typedef iList<iIdData> iPairList;

	//     
	typedef iListItem<iIdData> iIdItem;
	
	//  
	iIdString() : mIdTotal(0) {
		mpString = new iString;
	}

	//  
	~iIdString() {
		if(mpString) { delete mpString; mpString = NULL; }
	}

	//  
	iIdString(iIdString<iIdentifierType, iIdentifierItem>& aIdStr) {
		//  "" 
		mIdTotal = aIdStr.mIdTotal; aIdStr.mIdTotal = 0;
		mpString = aIdStr.mpString; aIdStr.mpString = NULL;
	}
	
	//   
	const iIdString& operator=(const iIdString& aIdStr) {
		// Make data copy
		mIdTotal  = aIdStr.mIdTotal;
		*mpString = *aIdStr.mpString;
		return *this;
	}

	//   +=
	const iIdString& operator+=(const iIdentifierList& aIdList) {
		Create(aIdList, true);
		return *this;
	}

	//    double
	operator double() const { return Total(); }

	//   
	void Create(const iIdentifierList& aIdList, bool aFlagAdd = false) {
		long count = 1;
		iIdentifierItem *actual, *prev = NULL;

		iString::iIfKeyFound pIfKeyFound = NULL;

		if(aFlagAdd) {
			//       
			pIfKeyFound = doIfKeyFound;
			mIdTotal   += aIdList.Total();
		} else {
			//  
			if(mpString->Total())  { mpString->Clear(); }
			mIdTotal = aIdList.Total();
		}

		iIdentifierList::iListItem* pIL = aIdList.GetFirstPosition();

		while(actual = iIdentifierList::GetNext(&pIL)) {
			if(prev) {
				//  ?
				if(actual->Identifier != prev->Identifier) {// 
					mpString->AddPair(prev->Identifier, count, pIfKeyFound);
					count = 1;
				} else { count++; }
			}
			prev = actual;
		}
		if(prev) { mpString->AddPair(prev->Identifier, count, pIfKeyFound); }
	}

	//    
	static double Similarity(const iIdString<iIdentifierType, iIdentifierItem>& aIdStr1,
							 const iIdString<iIdentifierType, iIdentifierItem>& aIdStr2) {
		
		double IdTotalMax  = MAX(aIdStr1.mIdTotal, aIdStr2.mIdTotal);

		if(!IdTotalMax) { return 1.0; }

		//        ? 
		bool IdStr1IsBiger = aIdStr1.Total() > aIdStr2.Total();
		
		double IdCoincidence = 0;
		iIdItem* pPair = ((IdStr1IsBiger)?aIdStr2:aIdStr1).GetFirstPosition();
		iIdData* pActualId;
		//  
		while(pActualId = GetNext(&pPair)) {
			long* pCount = ((IdStr1IsBiger)?aIdStr1:aIdStr2).mpString->Find(pActualId->Key);
			//   ?
			if(pCount) { // 
				IdCoincidence += MIN(*pCount, pActualId->Data);
			}
		}

		return IdCoincidence/IdTotalMax;
	}

	//  "  "
	static void doIfKeyFound(iPairList* apPL, const iIdData& aData, void* args) {
		//   
		iIdData* pActualData = apPL->Read();
		//   
		if(pActualData) {
			pActualData->Data += aData.Data;
		}
	}

	//      
	iIdItem* GetFirstPosition() const { return mpString->GetFirstPosition(); }

	//     
	static iIdData* GetNext(iIdItem** appLI) {
		return iString::iPairList::GetNext(appLI);
	}

	//      
	int Total() const { return mpString->Total(); }

	//      
	iString* GetStringPtr() const { return mpString; }
};

//      iIdString
template <class iIdentifierType, class iIdentifierItem>
bool operator==(const iIdString<iIdentifierType, iIdentifierItem>& s1,
				const iIdString<iIdentifierType, iIdentifierItem>& s2)
{
	return *s1.GetStringPtr() == *s2.GetStringPtr();
}

//      iIdString
template <class iIdentifierType, class iIdentifierItem>
bool operator>(const iIdString<iIdentifierType, iIdentifierItem>& s1,
				const iIdString<iIdentifierType, iIdentifierItem>& s2)
{
	return *s1.GetStringPtr() > *s2.GetStringPtr();
}

//      iIdString
template <class iIdentifierType, class iIdentifierItem>
bool operator<(const iIdString<iIdentifierType, iIdentifierItem>& s1,
				const iIdString<iIdentifierType, iIdentifierItem>& s2)
{
	return *s1.GetStringPtr() < *s2.GetStringPtr();
}


#endif // !defined(_ListInterface_H_INCLUDED_)